package lc.esp.script.engine;

import com.oracle.truffle.js.scriptengine.GraalJSScriptEngine;
import lc.mecha.log.MechaLogger;
import lc.mecha.log.MechaLoggerFactory;
import org.graalvm.polyglot.Context;

import javax.script.ScriptEngine;
import java.time.Duration;
import java.util.concurrent.*;

/**
 * A context for the ESP/SCRIPT system. This provides Javascript execution support within the ESP environment and
 * provides the necessary features for binding the ESP language features into the script environment.
 *
 * Security: The script engine provides support for partitioning the script from the host; for example, removing any
 * access to unsanctioned Java objects within the JVM as well as unsanctioned access to host resources list the
 * filesystem or networking. Basic controls are provided to limit overall script execution time.
 *
 * It is not thought that this implementation provides enough security control to allow the execution of mutually
 * un-trusted scripts. Only mutually trusted scripts should be executed within a single JVM task. This is due to a wide
 * variety of factors, not least of which are threats posed by spectre/meltdown type attacks. Protecting against that
 * class of attack is beyond the scope of this implementation, and, possibly, Java itself.
 *
 * @author Alex Leigh
 * @since mk20 (TOKYO PROTOCOL)
 */
public class ESPScriptEngine {
    private final ScriptEngine engine;
    private static final MechaLogger logger = MechaLoggerFactory.getLogger(ESPScriptEngine.class);
    private int timeoutMs = 10 * 1000;

    public ESPScriptEngine() {
        engine = GraalJSScriptEngine.create(null,
                Context.newBuilder("js")
                        .option("js.ecmascript-version", "2022")
                        .option("js.script-engine-global-scope-import", "false")
        );
    }

    public void eval(String script) throws TimeoutException, ExecutionException, InterruptedException {
        final ExecutorService executor = Executors.newSingleThreadExecutor();
        final Context context = Context.newBuilder("js").build();
        final Future<Object> futureResult = executor.submit(() -> context.eval("js", script));
        try {
            final Object result = futureResult.get(timeoutMs, TimeUnit.MILLISECONDS);
            logger.info("Script evaluated within 10 seconds, result: " + result);
        } catch (TimeoutException e) {
            context.interrupt(Duration.ZERO);
            logger.warn("Script execution time limit exceeded. Aborting.");
            throw new TimeoutException("Execution limit exceeded.");
        }
    }

    public int getTimeoutMs() {
        return timeoutMs;
    }

    public void setTimeoutMs(int timeoutMs) {
        this.timeoutMs = timeoutMs;
    }

    public Object get(String key) {
        return engine.get(key);
    }

    public ScriptEngine getEngine() {
        return engine;
    }
}
